# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.primitive.one_state import Counter, CreateJob, ExternalHamiltonian, GradientDescent, \
    InitialPositions, NEBForces
from pyiron_contrib.protocol.primitive.two_state import IsGEq
from pyiron_contrib.protocol.list import SerialList, AutoList
from pyiron_contrib.protocol.utils import Pointer

"""
Protocol for nudged elastic band (NEB) minimization.
"""

__author__ = "Liam Huber, Raynol Dsouza, Jan Janssen"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "18 July, 2019"


class NEB(CompoundVertex):
    """
    Relaxes a system according to the nudged elastic band method (Jonsson et al).

    Input attributes:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure_initial (Atoms): The starting structure for the elastic band.
        structure_final (Atoms): The final structure for the elastic band. Warning: must have the same number and
            species of atoms, cell, and periodic boundary conditions as the initial structure.
        n_images (int): How many images create by interpolating between the two structures.
        n_steps (int): How many steps to run for. (Default is 100.)
        f_tol (float): Ionic force convergence (largest atomic force). (Default is 1e-4 eV/angstrom.)
        spring_constant (float): Spring force between atoms in adjacent images. (Default is 1.0 eV/angstrom^2.)
        tangent_style ('plain'/'improved'/'upwinding'): How to calculate the image tangent in 3N-dimensional space.
            (Default is 'upwinding'.)
        use_climbing_image (bool): Whether to replace the force with one that climbs along the tangent direction for
            the job with the highest energy. (Default is True.)
        smoothing (float): Strength of the smoothing spring when consecutive images form an angle. (Default is None,
            do not apply such a force.)
        gamma0 (float): Initial step size as a multiple of the force. (Default is 0.1.)
        fix_com (bool): Whether the center of mass motion should be subtracted off of the position update. (Default
            is True)
        use_adagrad (bool): Whether to have the step size decay according to adagrad. (Default is False)

    Output attributes:
        energy_pot (list[float]): Total potential energy of the system in eV.
        positions (list[numpy.ndarray]): Atomic positions in angstroms for each image.
        forces (list[numpy.ndarray]): Atomic forces in eV/angstrom for each image, including spring forces.

    """

    def __init__(self, **kwargs):
        super(NEB, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.n_steps = 100
        id_.f_tol = 1e-4
        id_.spring_constant = 1.
        id_.tangent_style = 'upwinding'
        id_.use_climbing_image = True
        id_.smoothing = None
        id_.gamma0 = 0.1
        id_.fix_com = True
        id_.use_adagrad = False

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initialize_jobs = CreateJob()
        g.interpolate_images = InitialPositions()
        g.check_steps = IsGEq()
        g.calc_static = AutoList(ExternalHamiltonian)
        g.neb_forces = NEBForces()
        g.gradient_descent = SerialList(GradientDescent)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initialize_jobs,
            g.interpolate_images,
            g.check_steps, 'false',
            g.calc_static,
            g.neb_forces,
            g.gradient_descent,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = self.graph.initialize_jobs
        g.restarting_vertex = self.graph.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initialize_jobs
        g.initialize_jobs.input.n_images = ip.n_images
        g.initialize_jobs.input.ref_job_full_path = ip.ref_job_full_path
        g.initialize_jobs.input.structure = ip.structure_initial

        # interpolate_images
        g.interpolate_images.input.structure_initial = ip.structure_initial
        g.interpolate_images.input.structure_final = ip.structure_final
        g.interpolate_images.input.n_images = ip.n_images

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # calc_static
        g.calc_static.input.n_children = ip.n_images
        g.calc_static.direct.structure = ip.structure_initial
        g.calc_static.broadcast.project_path = gp.initialize_jobs.output.project_path[-1]
        g.calc_static.broadcast.job_name = gp.initialize_jobs.output.job_names[-1]
        g.calc_static.broadcast.default.positions = gp.interpolate_images.output.initial_positions[-1]
        g.calc_static.broadcast.positions = gp.gradient_descent.output.positions[-1]

        # neb_forces
        g.neb_forces.input.default.positions = gp.interpolate_images.output.initial_positions[-1]

        g.neb_forces.input.positions = gp.gradient_descent.output.positions[-1]
        g.neb_forces.input.energies = gp.calc_static.output.energy_pot[-1]
        g.neb_forces.input.forces = gp.calc_static.output.forces[-1]
        g.neb_forces.input.structure = ip.structure_initial
        g.neb_forces.input.spring_constant = ip.spring_constant
        g.neb_forces.input.tangent_style = ip.tangent_style
        g.neb_forces.input.use_climbing_image = ip.use_climbing_image
        g.neb_forces.input.smoothing = ip.smoothing

        # gradient_descent
        g.gradient_descent.input.n_children = ip.n_images
        g.gradient_descent.broadcast.default.positions = gp.interpolate_images.output.initial_positions[-1]

        g.gradient_descent.broadcast.positions = gp.gradient_descent.output.positions[-1]
        g.gradient_descent.broadcast.forces = gp.neb_forces.output.forces[-1]
        g.gradient_descent.direct.masses = ip.structure_initial.get_masses
        g.gradient_descent.direct.gamma0 = ip.gamma0
        g.gradient_descent.direct.fix_com = ip.fix_com
        g.gradient_descent.direct.use_adagrad = ip.use_adagrad

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_static.output.energy_pot[-1],
            'positions': ~gp.gradient_descent.output.positions[-1],
            'forces': ~gp.neb_forces.output.forces[-1]
        }

    def _get_energies(self, frame=None):
        if frame is None:
            return self.graph.calc_static.output.energy_pot[-1]
        else:
            return self.graph.calc_static.archive.output.energy_pot.data[frame]

    def plot_elastic_band(self, ax=None, frame=None, plot_kwargs=None):
        """
        Plot the string at an input frame. Here, frame is a dump of a step in the run. If `neb_job´ is the name
            of the neb job, the number of dumps can be specified by the user while submitting the job, as:

        >>> neb_job.set_output_whitelist(**{'calc_static': {'energy_pot': 20}})

        and run the job. Here, it dumps (or records a frame) of `energy_pot´ from the `calc_static´ vertex
            once every 20 steps.

        Default is plot the string at the final frame, as only the final dump is recorded. (unless specified
            otherwise by the user!)
        """
        if ax is None:
            _, ax = plt.subplots()
        if plot_kwargs is None:
            plot_kwargs = {}
        if 'marker' not in plot_kwargs.keys():
            plot_kwargs = {'marker': 'o'}
        energies = np.array(self._get_energies(frame=frame))
        ax.plot(energies - energies[0], **plot_kwargs)
        ax.set_ylabel("Energy")
        ax.set_xlabel("Image")
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        return ax

    def _get_directional_barrier(self, frame=None, anchor_element=0, use_minima=False):
        energies = self._get_energies(frame=frame)
        if use_minima:
            reference = energies.min()
        else:
            reference = energies[anchor_element]
        return energies.max() - reference

    def get_forward_barrier(self, frame=None, use_minima=False):
        """
        Get the energy barrier from the 0th image to the highest energy (saddle state).

        Args:
            frame (int): A particular dump. (Default is None, the final dump.)
            use_minima (bool): Whether to use the minima of the energies to compute the barrier. (Default is
                False, use the 0th value.)

        Returns:
            (float): the forward migration barrier.
        """
        return self._get_directional_barrier(frame=frame, use_minima=use_minima)

    def get_reverse_barrier(self, frame=None, use_minima=False):
        """
        Get the energy barrier from the final image to the highest energy (saddle state).

        Args:
            frame (int): A particular dump. (Default is None, the final dump.)
            use_minima (bool): Whether to use the minima of the energies to compute the barrier. (Default is
                False, use the nth value.)

        Returns:
            (float): the backward migration barrier.
        """
        return self._get_directional_barrier(frame=frame, anchor_element=-1, use_minima=use_minima)

    def get_barrier(self, frame=None, use_minima=True):
        return self.get_forward_barrier(frame=frame, use_minima=use_minima)
    get_barrier.__doc__ = get_forward_barrier.__doc__


class ProtoNEBSer(Protocol, NEB):
    pass



from __future__ import print_function
# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import inspect
from pkgutil import iter_modules

import numpy as np
import skimage
from matplotlib import pyplot as plt
from skimage import io
from collections import UserDict

from pyiron_contrib.image.utils import ModuleScraper

"""
Code for storing images in hdf5 and leveraging the skimage library as a class attribute.
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Feb 6, 2020"

# Some decorators look at the signature of skimage methods to see if they take an image
# (presumed to be in numpy.ndarray format).
# This is done by searching the signature for the variable name below:
_IMAGE_VARIABLE = 'image'


def pass_image_data(image):
    """
    Decorator to see if the signature of the function starts with a particular variable (`_IMAGE_VARIABLE`). If so,
    automatically passes an attribute of the argument (`image.data`) as the first argument.

    Args:
        image (Image): The image whose data to use.

    Returns:
        (fnc): Decorated function.
    """

    def decorator(function):
        takes_image_data = list(inspect.signature(function).parameters.keys())[0] == _IMAGE_VARIABLE

        def wrapper(*args, **kwargs):
            if takes_image_data:
                return function(image.data, *args, **kwargs)
            else:
                return function(*args, **kwargs)

        wrapper.__doc__ = ""
        if takes_image_data:
            wrapper.__doc__ += "This function has been wrapped to automatically supply the image argument. \n" \
                               "Remaining arguments can be passed as normal.\n"
        wrapper.__doc__ += "The original docstring follows:\n\n"
        wrapper.__doc__ += function.__doc__ or ""
        return wrapper

    return decorator


def set_image_data(image):
    """
    Decorator which checks the returned value of the function. If that value is of type `numpy.ndarray`, uses it to set
    an attribute of the argument (`image.data`) instead of returning it.

    Args:
        image (Image): The image whose data to set.

    Returns:
        (fnc): Decorated function.
    """

    def decorator(function):
        def wrapper(*args, **kwargs):
            output = function(*args, **kwargs)
            if isinstance(output, np.ndarray):
                image._data = output
            else:
                return output

        wrapper.__doc__ = "This function has been wrapped; if it outputs a numpy array, it will be " \
                          "automatically passed to the image's data field.\n" + function.__doc__
        return wrapper

    return decorator


def pass_and_set_image_data(image):
    """
    Decorator which connects function input and output to `image.data`.

    Args:
        image (Image): The image whose data to set.

    Returns:
        (fnc): Decorated function.
    """

    def decorator(function):
        return set_image_data(image)(pass_image_data(image)(function))
    return decorator


class Image:
    """
    A base class for storing image data in the form of numpy arrays. Functionality of the skimage library can be
    leveraged using the sub-module name and an `activate` method.

    Attributes:
        source (str/numpy.ndarray): The raw data source.
        data (numpy.ndarray): The image data. Not loaded until called, so first call may be slow. Modifications are
            made to this field, leaving the source untouched. NOT saved to hdf.
        as_gray (bool): Whether to interpret the image as grayscale.
        metadata (Metadata): Metadata associated with the source.
    """

    def __init__(self, source, metadata=None, as_gray=False):
        """
        source (str/numpy.ndarray): The raw data source.
        as_gray (bool): Whether to interpret the image as grayscale. (Default is False)
        metadata (Metadata): Metadata associated with the source. (Default is None.)
        """

        # Set data
        self._source = source
        self._data = None
        self.as_gray = as_gray

        # Set metadata
        self._metadata = None
        self.metadata = metadata or Metadata()

        # Apply wrappers
        submodule_blacklist = [
            'data',
            'scripts',
            'future',
            'registration',

        ]
        for module in iter_modules(skimage.__path__):
            if module.name[0] == '_' or module.name in submodule_blacklist:
                continue
            setattr(
                self,
                module.name,
                ModuleScraper(
                    'skimage.' + module.name,
                    decorator=pass_and_set_image_data,
                    decorator_args=(self,)
                )
            )

    @property
    def source(self):
        return self._source

    def overwrite_source(self, new_source, new_metadata=None, as_gray=False):
        """
        Apply a new source of image data to the image object.

        Args:
            new_source (str/numpy.ndarray): The filepath to the data, or the raw array of data itself.
            new_metadata (Metadata): The metadata associated with the new source. (Default is None.)
            as_gray (bool): Whether to interpret the new data as grayscale. (Default is False.)
        """

        self._source = new_source
        self._data = None
        self.as_gray = as_gray
        self.metadata = new_metadata or Metadata()

    @property
    def data(self):
        if self._data is None:
            self._load_data_from_source()
        return self._data

    @property
    def shape(self):
        return self.data.shape

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, new_metadata):
        if new_metadata is None or isinstance(new_metadata, Metadata):
            self._metadata = new_metadata
        elif isinstance(new_metadata, dict):
            self._metadata = Metadata(new_metadata)
        else:
            raise ValueError("Metadata field expected a `dict`, `Metadata`, or `None`, but got {}".format(
                type(new_metadata))
            )

    def __len__(self):
        return self.data.__len__()

    def _load_data_from_source(self):
        if isinstance(self.source, np.ndarray):
            self._data = self.source.copy()
            if len(self._data.shape) == 3:
                self.convert_to_grayscale()
        elif isinstance(self.source, str):
            self._data = io.imread(self.source, as_gray=self.as_gray)
        else:
            raise ValueError("Data source not understood, should be numpy.ndarray or string pointing to image file.")

    def reload_data(self):
        """
        Reverts the `data` attribute to the source, i.e. the most recently read file (if set by reading data), or the
        originally assigned array (if set by direct array assignment).
        """

        self._load_data_from_source()

    def convert_to_grayscale(self):
        """
        Flattens (NxMx3) data into (NxM) grayscale data.
        """
        if self._data is not None:
            if len(self.data.shape) == 3 and self.data.shape[-1] == 3:
                self._data = np.mean(self._data, axis=-1)
                self.as_gray = True
            else:
                raise ValueError("Can only convert data with shape NxMx3 to grayscale")
        else:
            self.as_gray = True

    def plot(self, ax=None, subplots_kwargs=None, imshow_kwargs=None, hide_axes=True):
        """
        Make a simple matplotlib `imshow` plot of the data.

        Args:
            ax (matplotlib.axes.Axes): The axis to plot on. (Default is None, make a new figure.)
            subplots_kwargs (dict): Keyword arguments to pass to the figure generation. Only used if no axis is
                provided. (Default is None.)
            imshow_kwargs (dict): Keyword arguments to pass to the `imshow` plotting command. (Default is None.)
            hide_axes (bool): Whether to hide axis ticks and labels. (Default is True.)

        Returns:
            (matplotlib.figure.Figure): The figure the plot is in.
            (matplotlib.axes.Axes): The axis the plot is on.
        """

        subplots_kwargs = subplots_kwargs or {}
        imshow_kwargs = imshow_kwargs or {}

        if ax is None:
            fig, ax = plt.subplots(**subplots_kwargs)
        else:
            fig = ax.figure

        ax.imshow(self.data, **imshow_kwargs)

        if hide_axes:
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)

        return fig, ax

    def to_hdf(self, hdf, group_name=None):
        with hdf.open(group_name) as hdf5_server:
            hdf5_server["TYPE"] = str(type(self))
            hdf5_server["source"] = self.source
            hdf5_server["as_gray"] = self.as_gray
            self.metadata.to_hdf(hdf=hdf5_server, group_name="metadata")

    def from_hdf(self, hdf, group_name=None):
        with hdf.open(group_name) as hdf5_server:
            source = hdf5_server["source"]
            as_gray = hdf5_server["as_gray"]
            metadata = Metadata()
            metadata.from_hdf(hdf=hdf5_server, group_name="metadata")
        self.overwrite_source(source, new_metadata=metadata, as_gray=as_gray)


class Metadata(UserDict):
    """
    TODO: Leverage the generic to and from hdf functions written by Dominik over in pyiron_contrib/protocol
    """

    def __getattr__(self, item):
        return self.data[item]
        # return super(Metadata, self).__getitem__(item)

    def __setattr__(self, key, value):
        if key == "data":
            self.__dict__[key] = value
        else:
            self.__dict__['data'][key] = value

    def to_hdf(self, hdf, group_name=None):
        with hdf.open(group_name) as hdf5_server:
            hdf5_server["TYPE"] = str(type(self))
            hdf5_server["KEYS"] = list(self.keys())
            for k, v in self.items():
                hdf5_server[k] = v

    def from_hdf(self, hdf, group_name=None):
        with hdf.open(group_name) as hdf5_server:
            keys = hdf5_server["KEYS"]
            for k in keys:
                self[k] = hdf5_server[k]



__version__ = "0.1"
__all__ = []

import warnings

try:
	from pyiron import Project
except:
	warnings.warn("pyiron module not found, importing Project from pyiron_base")
	from pyiron_base import Project

from pyiron_base import JOB_CLASS_DICT

# Make classes available for new pyiron version
JOB_CLASS_DICT['ProtoMinimGradDes'] = 'pyiron_contrib.protocol.compound.minimize'
JOB_CLASS_DICT['ProtoMD'] = 'pyiron_contrib.protocol.compound.molecular_dynamics'
JOB_CLASS_DICT['ProtoConfinedMD'] = 'pyiron_contrib.protocol.compound.molecular_dynamics'
JOB_CLASS_DICT['ProtoHarmMD'] = 'pyiron_contrib.protocol.compound.molecular_dynamics'
JOB_CLASS_DICT['ProtoNEBSer'] = 'pyiron_contrib.protocol.compound.nudged_elastic_band'
JOB_CLASS_DICT['ProtocolQMMM'] = 'pyiron_contrib.protocol.compound.qmmm'
JOB_CLASS_DICT['ProtoHarmTILDSer'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoHarmTILDPar'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoVacTILDSer'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoVacTILDPar'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoVacForm'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoFTSEvoSer'] = 'pyiron_contrib.protocol.compound.finite_temperature_string'
JOB_CLASS_DICT['ProtoFTSEvoPar'] = 'pyiron_contrib.protocol.compound.finite_temperature_string'
JOB_CLASS_DICT['ImageJob'] = 'pyiron_contrib.image.job'
JOB_CLASS_DICT['LangevinAse'] = 'pyiron_contrib.atomistics.interactive.langevin'
JOB_CLASS_DICT['Mixer'] = 'pyiron_contrib.atomistics.interactive.mixer'
JOB_CLASS_DICT['ParameterMaster'] = 'pyiron_contrib.atomistics.dft.parametermaster'
JOB_CLASS_DICT['MonteCarloMaster'] = 'pyiron_contrib.atomistics.interactive.montecarlo'
JOB_CLASS_DICT['RandSpg'] = 'pyiron_contrib.atomistics.randspg.randspg'
JOB_CLASS_DICT['Fenics'] = 'pyiron_contrib.continuum.fenics.job.generic'
JOB_CLASS_DICT['FenicsLinearElastic'] = 'pyiron_contrib.continuum.fenics.job.elastic'
JOB_CLASS_DICT['TrainingContainer'] = 'pyiron_contrib.atomistics.atomistics.job.trainingcontainer'
JOB_CLASS_DICT['RandomDisMaster'] = 'pyiron_contrib.atomistics.mlip.masters'
JOB_CLASS_DICT['RandomMDMaster'] = 'pyiron_contrib.atomistics.mlip.masters'
JOB_CLASS_DICT['RunnerFit'] = 'pyiron_contrib.atomistics.runner.job'
JOB_CLASS_DICT['MlipSelect'] = 'pyiron_contrib.atomistics.mlip.mlipselect'
JOB_CLASS_DICT['Mlip'] = 'pyiron_contrib.atomistics.mlip.mlip'
JOB_CLASS_DICT['LammpsMlip'] = 'pyiron_contrib.atomistics.mlip.lammps'
JOB_CLASS_DICT['MlipJob'] = 'pyiron_contrib.atomistics.mlip.mlipjob'
JOB_CLASS_DICT['Atomicrex'] = 'pyiron_contrib.atomistics.atomicrex.atomicrex_job'


from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from ase import units
from ase.md.langevin import Langevin
from pyiron_base import GenericParameters
from pyiron.gpaw.pyiron_ase import AseAdapter
from pyiron.atomistics.job.interactivewrapper import InteractiveWrapper

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class LangevinAse(InteractiveWrapper):
    """
    Reference implementation of a thermostat implemented in ASE. While we have multiple implementations of the
    Langevin thermostat already, this one is using the ASE interface and the Langevin thermostat implemented in ASE.
    It is therefore a perfect example how other thermostats implemented in ASE can be used inside pyiron.

    Args:
        project (ProjectHDFio): points to the HDF5 file the job is stored in
        job_name (str): name of the job, which has to be unique within the project

    Attributes:
        input (pyiron.objects.hamilton.md.lammps.Input): handle the input
    """

    def __init__(self, project, job_name):
        super(LangevinAse, self).__init__(project, job_name)
        self.__name__ = "LangevinAse"
        self.input = Input()
        self._fast_mode = False

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        super(LangevinAse, self).set_input_to_read_only()
        self.input.read_only = True

    def write_input(self):
        """
        No input is written when using the ASE interface
        """
        pass

    def _write_run_wrapper(self, debug=False):
        """
        No wrapper is written for the ASE interface

        Args:
            debug (bool): the debug flag is ignored
        """
        pass

    def run_static(self):
        """
        Static run function, which uses the ASEAdapter to connect the pyiron interactive reference job with the
        Langevin thermostat implemented in ASE, by setting the ASEAdapter as a replacement of the ASE atoms object.
        """
        self.status.running = True
        self.ref_job_initialize()
        aseadapter = AseAdapter(
            self.ref_job,
            self._fast_mode
        )
        langevin = Langevin(
            atoms=aseadapter,
            timestep=self.input['time_step'] * units.fs,
            temperature=self.input['temperature'] * units.kB,
            friction=self.input['friction'],
            fixcm=True
        )
        langevin.run(self.input['ionic_steps'])
        self.status.collect = True
        aseadapter.interactive_close()
        self._finish_job()


class Input(GenericParameters):
    """
    class to control the generic input for a Sphinx calculation.

    Args:
        input_file_name (str): name of the input file
        table_name (str): name of the GenericParameters table
    """

    def __init__(self, input_file_name=None, table_name="input"):
        super(Input, self).__init__(input_file_name=input_file_name, table_name=table_name, comment_char="#",
                                    separator_char="=", end_value_char=';')

    def load_default(self):
        """
        Loads the default file content
        """
        file_content = ('ionic_steps = 100\n'
                        'temperature = 1500\n'
                        'time_step = 1\n'
                        'friction = 0.002\n'
                        'fix_center_of_mass = True\n')
        self.load_string(file_content)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

import os
import posixpath

from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_base import GenericParameters, GenericJob
from pyiron_atomistics.vasp.structure import read_atoms

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class RandSpg(GenericJob):
    """
    RandSpg is a program that generates random crystals with specific space groups.
    The user inputs a specific composition and space group to be generated. The
    algorithm then proceeds by finding all combinations of Wyckoff positions that
    satisfy the composition. It then randomly selects a combination of Wyckoff
    positions for the space group, generates random coordinates for variables
    in the Wyckoff positions, and places atoms in those sites. It ensures that any
    constraints the user placed on the system (lattice constraints including
    min/max volume, minimum interatomic distances, specific Wyckoff positions for
    specific atoms, etc.) are satisfied.
    Code: https://github.com/xtalopt/randSpg
    Paper: https://www.sciencedirect.com/science/article/pii/S0010465516303848
    """
    def __init__(self, project, job_name):
        super(RandSpg, self).__init__(project, job_name)
        self.__version__ = "0.1"
        self.__name__ = "RandSpg"
        self._lst_of_struct = []
        self.input = ExampleInput()
        self._executable_activate()

    @property
    def list_of_structures(self):
        return self._lst_of_struct

    def list_structures(self):
        if self.status.finished:
            return self._lst_of_struct
        else:
            return []

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        self.input.read_only = True

    # define routines that create all necessary input files
    def write_input(self):
        """
        Call routines that generate the codespecifc input files
        """
        self.input.write_file(file_name="randspg.in", cwd=self.working_directory)

    # define routines that collect all output files
    def collect_output(self):
        """
        Parse the output files of the example job and store the results in the HDF5 File.
        """
        self.collect_output_log()

    def collect_output_log(self, dir_name="randSpgOut"):
        """
        general purpose routine to extract output from logfile

        Args:
            file_name (str): output.log - optional
        """
        self._lst_of_struct = [[file_name.replace('-', '_'),
                                read_atoms(filename=posixpath.join(self.working_directory, dir_name, file_name))]
                               for file_name in os.listdir(posixpath.join(self.working_directory, dir_name))]
        for structure_name, structure in self._lst_of_struct:
            with self.project_hdf5.open("output/structures/" + structure_name) as h5:
                structure.to_hdf(h5)

    def collect_logfiles(self):
        pass

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the ExampleJob object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(RandSpg, self).to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            self.input.to_hdf(hdf5_input)

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the ExampleJob object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(RandSpg, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            self.input.from_hdf(hdf5_input)
        self._lst_of_struct = []
        with self.project_hdf5.open("output/structures") as hdf5_output:
            structure_names = hdf5_output.list_groups()
        for group in structure_names:
            with self.project_hdf5.open("output/structures/" + group) as hdf5_output:
                self._lst_of_struct.append([group, Atoms().from_hdf(hdf5_output)])


class ExampleInput(GenericParameters):
    def __init__(self, input_file_name=None):
        super(ExampleInput, self).__init__(input_file_name=input_file_name,
                                           table_name="randspg_in",
                                           comment_char="#",
                                           separator_char='=')

    def load_default(self):
        """
        Loading the default settings for the input file.
        """
        input_str = '''\
#Inputfile
composition = Mg4Al2
spacegroups = 1-8, 10, 25, 28, 30-32
latticeMins =  3.0,  3.0,  3.0,  60.0,  60.0,  60.0
latticeMaxes = 10.0, 10.0, 10.0, 120.0, 120.0, 120.0
minVolume = 450
maxVolume = 500
numOfEachSpgToGenerate = 3
setMinRadii = 0.3
scalingFactor = 0.5
maxAttempts = 100
outputDir = randSpgOut
verbosity = r
'''
        self.load_string(input_str)

    def write_file(self, file_name, cwd=None):
        """
        Write GenericParameters to input file

        Args:
            file_name (str): name of the file, either absolute (then cwd must be None) or relative
            cwd (str): path name (default: None)
        """
        if cwd is not None:
            file_name = posixpath.join(cwd, file_name)

        with open(file_name, "w") as f:
            for line in self.get_string_lst():
                f.write(line.replace("(", "").replace(")", ""))



from pyiron_base import DataContainer

class Output(DataContainer):
    """
    Class to store general output quantities.
    Final properties and function parameter values are stored within
    the respective classes and not here.
    """    
    def __init__(self, *args, **kwargs):
        super().__init__(table_name="output", *args, **kwargs)
        self.error = None
        self.residual = None
        self.iterations = None


# coding: utf-8
# http://gitlab.skoltech.ru/shapeev/mlip-dev/blob/master/src/external/python/mlippy/cfgs.py

from __future__ import print_function
import numpy as np


class Cfg:
    pos = None
    lat = None
    types = None
    energy = None
    forces = None
    stresses = None
    desc = None
    grade = None


def readcfg(f):
    cfg = Cfg()
    cfg.lat = np.zeros((3, 3))
    size = -1
    mode = -1
    line = f.readline()
    while line:
        line = line.upper()
        line = line.strip()
        if mode == 0:
            if line.startswith('SIZE'):
                line = f.readline()
                size = int(line.strip())
                cfg.types = np.zeros(size)
                cfg.pos = np.zeros((size, 3))
            elif line.startswith('SUPERCELL'):
                line = f.readline()
                vals = line.strip().split()
                cfg.lat[0, :] = vals[0:3]
                line = f.readline()
                vals = line.strip().split()
                cfg.lat[1, :] = vals[0:3]
                line = f.readline()
                vals = line.strip().split()
                cfg.lat[2, :] = vals[0:3]
            elif line.startswith('ATOMDATA'):
                if line.endswith('FZ'):
                    cfg.forces = np.zeros((size, 3))
                for i in range(size):
                    line = f.readline()
                    vals = line.strip().split()
                    cfg.types[i] = vals[1]
                    cfg.pos[i, :] = vals[2:5]
                    if cfg.forces is not None:
                        cfg.forces[i, :] = vals[5:8]
            elif line.startswith('ENERGY'):
                line = f.readline()
                cfg.energy = float(line.strip())
            elif line.startswith('PLUSSTRESS'):
                line = f.readline()
                vals = line.strip().split()
                cfg.stresses = np.zeros(6)
                cfg.stresses[:] = vals[0:6]
            elif line.startswith('FEATURE   MV_GRADE'):
                cfg.grade = float(line.split()[-1])
            elif line.startswith('FEATURE   PYIRON'):
                cfg.desc = line.split()[-1]
        if line.startswith('BEGIN_CFG'):
            mode = 0
        elif line.startswith('END_CFG'):
            break
        line = f.readline()
    return cfg


def savecfg(f, cfg, desc=None):
    atstr1 = 'AtomData:  id type       cartes_x      cartes_y      cartes_z           fx          fy          fz'
    atstr2 = 'AtomData:  id type       cartes_x      cartes_y      cartes_z'
    size = len(cfg.types)
    print('BEGIN_CFG', file=f)
    print('Size', file=f)
    print('   %-d' % size, file=f)
    if cfg.lat is not None:
        print('SuperCell', file=f)
        for i in range(3):
            print('         %14f%14f%14f'
                  % (cfg.lat[i, 0], cfg.lat[i, 1], cfg.lat[i, 2]), file=f)
    if cfg.forces is not None:
        print(atstr1, file=f)
    else:
        print(atstr2, file=f)
    for i in range(size):
        if cfg.forces is not None:
            print('         %4d %4d %14f%14f%14f %16.8e %16.8e %16.8e' %
                  (i+1, cfg.types[i], cfg.pos[i, 0], cfg.pos[i, 1], cfg.pos[i, 2],
                   cfg.forces[i, 0], cfg.forces[i, 1], cfg.forces[i, 2]), file=f)
        else:
            print('         %4d %4d %14f%14f%14f' %
                  (i+1, cfg.types[i], cfg.pos[i, 0], cfg.pos[i, 1], cfg.pos[i, 2]),
                  file=f)
    if cfg.energy is not None:
        print('Energy\t%14f' % cfg.energy, file=f)
    if cfg.stresses is not None:
        print('PlusStress:  xx          yy          zz          yz          xz          xy', file=f)
        print('         %14f%14f%14f%14f%14f%14f' %
              (cfg.stresses[0], cfg.stresses[1], cfg.stresses[2],
               cfg.stresses[3], cfg.stresses[4], cfg.stresses[5]), file=f)
    if desc is not None:
        print('Feature   from %s' % desc, file=f)
    if cfg.desc is not None:
        print('Feature %s' % cfg.desc, file=f)
    print('END_CFG', file=f)


class cfgparser:
    def __init__(self, file, max_cfgs=None):
        self.cfgs = []
        self.file = file
        self.max_cfgs = max_cfgs

    def __enter__(self):
        while True:
            if self.max_cfgs is not None and len(self.cfgs) == self.max_cfgs:
                break
            cfg = readcfg(self.file)
            if cfg.types is not None:
                self.cfgs.append(cfg)
            else:
                break
        return self.cfgs

    def __exit__(self, *args):
        self.cfgs = []


def printcfg(cfg):
    savecfg(None, cfg)


def loadcfgs(filename, max_cfgs=None):
    with open(filename, 'r') as file:
        with cfgparser(file, max_cfgs) as cfgs:
            return cfgs


def savecfgs(filename, cfgs, desc=None):
    with open(filename, 'w') as file:
        for cfg in cfgs:
            savecfg(file, cfg, desc)
            print("", file=file)



import xml.etree.ElementTree as ET
import posixpath

import numpy as np
from ase.data import atomic_masses, atomic_numbers

from pyiron_base import DataContainer, PyironFactory

from pyiron_contrib.atomistics.atomicrex.utility_functions import write_pretty_xml
from pyiron_contrib.atomistics.atomicrex.potential_factory import ARPotFactory
from pyiron_contrib.atomistics.atomicrex.function_factory import FunctionFactory

class GeneralARInput(DataContainer):
    """
    Class to store general input of an atomicrex job,
    f.e. the fit algorithm.
    """  
    def __init__(
        self,
        table_name="general_input",
        name="Atomicrex Job",
        verbosity="medium",
        real_precision = 16,
        valdiate_potentials = False,
        atom_types = None,
        fit_algorithm = None,
        output_interval = 10000,
        enable_fitting = True,
        ):
        
        super().__init__(table_name="general_input",)
        
        self.name = name
        self.verbosity = verbosity
        self.real_precision = real_precision
        self.validate_potentials = valdiate_potentials
        self.atom_types = AtomTypes()      
        self.fit_algorithm = fit_algorithm
        self.output_interval = output_interval
        self.enable_fitting = enable_fitting
        

    def _write_xml_file(self, directory):
        """Internal function.
        Write the main input xml file in a directory.   

        Args:
            directory (str): Working directory
        """        
        job = ET.Element("job")

        name = ET.SubElement(job, "name")
        name.text = self.name

        verbosity = ET.SubElement(job, "verbosity")
        verbosity.text = self.verbosity

        if self.validate_potentials:
            validate_potentials = ET.SubElement(job, "validate_potentials")

        real_precision = ET.SubElement(job, "real-precision")
        real_precision.text = f"{self.real_precision}"

        atom_types = ET.SubElement(job, "atom-types")
        for k, v in self.atom_types.items():
            species = ET.SubElement(atom_types, "species")
            species.text = k
            if v is not None:
                mass, index = v
            else:
                index = atomic_numbers[k]
                mass = atomic_masses[index]

            species.set("mass", f"{mass}")
            species.set("atomic-number", f"{index}")

        fitting = ET.SubElement(job, "fitting")
        if self.enable_fitting:
            fitting.set("enabled", "true")
        else:
            fitting.set("enabled", "false")
        fitting.set("output-interval", f"{self.output_interval}")
        fitting.append(self.fit_algorithm._to_xml_element())

        potentials = ET.SubElement(job, "potentials")
        include = ET.SubElement(potentials, "xi:include")
        include.set("href", "potential.xml")
        include.set("xmlns:xi", "http://www.w3.org/2003/XInclude")

        structures = ET.SubElement(job, "structures")
        include = ET.SubElement(structures, "xi:include")
        include.set("href", "structures.xml")
        include.set("xmlns:xi", "http://www.w3.org/2003/XInclude")

        file_name = posixpath.join(directory, "main.xml")
        write_pretty_xml(job, file_name)


class AtomTypes(DataContainer):
    """
        DataContainer used to specify elements in the fit job.
        Elements can be added using dictionary or attribute syntax.
        Their value should None or a (mass, index) tuple as value.
        If value is None the mass and index are taken from the ase package.
        Examples:
        job.input.atom_types.Cu = None
        job.input.atom_types["Cu"] = None 
        job.input.atom_types.Cu = (63.546, 29)
    """
    def __init__(self, *args, **kwargs):
        super().__init__(table_name="AtomTypes", *args, **kwargs)


class AlgorithmFactory(PyironFactory):
    """
    Factory class to conveniently acces the different
    fitting algorithms available in atomicrex.
    """    

    @staticmethod
    def ar_lbfgs(conv_threshold=1e-10, max_iter=50, gradient_epsilon=None):
        return AtomicrexAlgorithm(name="BFGS", conv_threshold=conv_threshold, max_iter=max_iter, gradient_epsilon=gradient_epsilon)

    @staticmethod
    def ar_spa(spa_iterations=20, seed=42):
        return SpaMinimizer(spa_iterations=spa_iterations, seed=seed)

    @staticmethod
    def ld_lbfgs(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LD_LBFGS", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def ld_mma(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LD_MMA", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)
    
    @staticmethod
    def ld_ccsaq(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LD_CCSAQ", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def ld_slsqp(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LD_SLSQP", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)
    
    @staticmethod
    def ld_var1(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LD_VAR1", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def ld_var2(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LD_VAR2", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def ln_cobyla(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LN_COBYLA", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def ln_bobyqa(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LN_BOBYQA", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)
    
    @staticmethod
    def ln_newuoa(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LN_NEWUOA", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)
    
    @staticmethod
    def ln_newuoa_bound(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LN_NEWUOA_BOUND", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)
    
    @staticmethod
    def ln_praxis(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LN_PRAXIS", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def ln_neldermead(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LN_NELDERMEAD", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def ln_sbplx(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="LN_SBPLX", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def gn_crs2_lm(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="GN_CRS2_LM", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def gn_esch(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="GN_ESCH", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def gn_direct(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="GN_DIRECT", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def gn_direct_l(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="GN_DIRECT_L", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)

    @staticmethod
    def gn_isres(stopval=1e-10, max_iter=50, maxtime=None, ftol_rel=None, ftol_abs=None, xtol_rel=None, seed=None):
        return NloptAlgorithm(name="GN_ISRES", stopval=stopval, max_iter=max_iter, maxtime=maxtime, ftol_rel=ftol_rel, ftol_abs=ftol_abs, xtol_rel=xtol_rel, seed=seed)


class AtomicrexAlgorithm(DataContainer):
    """
    Class to inherit from. If more algorithms will be implemented in atomicrex
    at some point they can be implemented similar to the AR_LBFGS class.
    """    
    def __init__(self, conv_threshold=None, max_iter=None, gradient_epsilon=None, name=None, *args, **kwargs):
        super().__init__(table_name="fitting_algorithm", *args, **kwargs)
        self.conv_threshold = conv_threshold
        self.max_iter = max_iter
        self.gradient_epsilon = gradient_epsilon
        self.name = name

    def _to_xml_element(self):
        """Internal function.
        Converts the algorithm to an xml element
        """        
        algo = ET.Element(self.name)
        algo.set("conv-threshold", f"{self.conv_threshold}")
        algo.set("max-iter", f"{self.max_iter}")
        if self.gradient_epsilon is not None:
            algo.set("gradient-epsilon", f"{self.gradient_epsilon}")
        return algo


class SpaMinimizer:
    """
    Global optimizer implemented in atomicrex.
    Should be used in combination with a local minimizer.
    See the atomicrex documentation for details.
    """    
    def __init__(self, spa_iterations, seed):
        self.spa_iterations = spa_iterations
        self.seed = seed
        self.local_minimizer = None

    def _to_xml_element(self):
        """Internal function.
        Converts the algorithm to a xml element
        and returns it
        """     
        spa = ET.Element("spa")
        spa.set("max-iter", f"{self.spa_iterations}")
        spa.set("seed", f"{self.seed}")
        if self.local_minimizer is not None:
            spa.append(self.local_minimizer._to_xml_element())
        return spa


class NloptAlgorithm(DataContainer):
    """
    Nlopt algorithms should inherit from this class.
    """    
    def __init__(
        self,
        name=None,
        seed=None,
        stopval=None,
        max_iter=None,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        *args, **kwargs):
        super().__init__(table_name="fitting_algorithm", *args, **kwargs)
        self.stopval = stopval
        self.max_iter = max_iter
        self.maxtime = maxtime
        self.ftol_rel = ftol_rel
        self.ftol_abs = ftol_abs
        self.xtol_rel = xtol_rel
        self.seed = seed
        self.name = name

    def _to_xml_element(self):
        """Internal Function.
        Converts the algorithm to a xml element
        and returns it
        """        
        nlopt = ET.Element("nlopt")
        nlopt.set("algorithm", self.name)
        if self.stopval is not None:
            nlopt.set("stopval", f"{self.stopval}")
        if self.max_iter is not None:
            nlopt.set("maxeval", f"{self.max_iter}")
        if self.maxtime is not None:
            nlopt.set("maxtime", f"{self.maxtime}")
        if self.ftol_rel is not None:
            nlopt.set("ftol_rel", f"{self.ftol_rel}")
        if self.ftol_abs is not None:
            nlopt.set("ftol_abs", f"{self.ftol_abs}")
        if self.xtol_rel is not None:
            nlopt.set("xtol_rel", f"{self.xtol_rel}")
        return nlopt


class NloptGlobalLocal(NloptAlgorithm):
    """
    Nlopt global optimizers that additionally need a local minimizer similar to the spa algorithm implemented in atomicrex
    """    
    def __init__(self, stopval, max_iter, maxtime, ftol_rel, ftol_abs, xtol_rel, name, seed, *args, **kwargs):
        super().__init__(
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            name=name,
            seed=seed,
            *args,
            **kwargs)
        self.local_minimizer = None

    def _to_xml_element(self):
        """Internal Function.
        Converts the algorithm to a xml element
        and returns it
        """ 
        if self.local_minimizer is None:
            raise ValueError("This global minimzer needs an additional local minimzer")
        nlopt = super()._to_xml_element()
        local = self.local_minimizer._to_xml_element()
        nlopt.append(local)
        return nlopt
        


import xml.etree.ElementTree as ET
from xml.dom import minidom

def write_pretty_xml(elem, filename):
    """
    The files directly written with xml.etree.ElemenTrees contain no newlines and indents. 
    This is a helper function to write them in a human readable way.

    Args:
        elem (ElementTree): xml.etree.ElementTree ElementTree object to be written
        filename (string): name for the xml file
    """    
    ugly = ET.tostring(elem, 'utf-8')
    reparsed = minidom.parseString(ugly)
    reparsed = reparsed.toprettyxml(indent="    ")
    with open(filename, "w") as f:
        f.write(reparsed)
    return


# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from pyiron_contrib.protocol.generic import PrimitiveVertex
import numpy as np
from abc import abstractmethod
from scipy.linalg import toeplitz

"""
Vertices whose present application extends only to finite temperature string-based protocols.
"""

__author__ = "Raynol Dsouza, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "20 July, 2019"


class _StringDistances(PrimitiveVertex):
    """
    A parent class for vertices which care about the distance from an image to various centroids on the string.
    """
    def __init__(self, name=None):
        super(PrimitiveVertex, self).__init__(name=name)
        self.input.default.eps = 1e-6

    @abstractmethod
    def command(self, *args, **kwargs):
        pass

    @staticmethod
    def check_closest_to_parent(structure, positions, centroid_positions, all_centroid_positions, eps):
        """
        Checks which centroid the image is closest too, then measures whether or not that closest centroid is sufficiently
            close to the image's parent centroid.
        Args:
            structure (Atoms): The reference structure.
            positions (numpy.ndarray): Atomic positions of this image.
            centroid_positions (numpy.ndarray): The positions of the image's centroid.
            all_centroid_positions (list/numpy.ndarray): A list of positions for all centroids in the string.
            eps (float): The maximum distance between the closest centroid and the parent centroid to be considered a match
                (i.e. no recentering necessary).
        Returns:
            (bool): Whether the image is closest to its own parent centroid.
        """
        distances = [np.linalg.norm(structure.find_mic(c_pos - positions)) for c_pos in all_centroid_positions]
        closest_centroid_positions = all_centroid_positions[np.argmin(distances)]
        match_distance = np.linalg.norm(structure.find_mic(closest_centroid_positions - centroid_positions))
        return match_distance < eps


class StringRecenter(_StringDistances):
    """
    If not, the image's positions and forces are reset to match its centroid.

    Input attributes:
        positions (numpy.ndarray): Atomic positions of the image.
        forces (numpy.ndarray): Atomic forces on the image.
        centroid_positions (numpy.ndarray): The positions of the image's centroid.
        centroid_forces (numpy.ndarray): The forces on the image's centroid.
        all_centroid_positions (list/numpy.ndarray): A list of positions for all centroids in the string.
        structure (Atoms): The reference structure.
        eps (float): The maximum distance between the closest centroid and the parent centroid to be considered a match
            (i.e. no recentering necessary). (Default is 1e-6.)

    Output attributes:
        positions (numpy.ndarray): Either the original positions passed in, or the centroid positions.
        forces (numpy.ndarray): Either the original forces passed in, or the centroid forces.
        recentered (bool): Whether or not the image got recentered.
    """
    def command(self, structure, positions, forces, centroid_positions, centroid_forces, all_centroid_positions, eps):
        if self.check_closest_to_parent(structure, positions, centroid_positions, all_centroid_positions, eps):
            return {
                'positions': positions,
                'forces': forces,
                'recentered': False,
            }
        else:
            return {
                'positions': centroid_positions,
                'forces': centroid_forces,
                'recentered': True
            }


class StringReflect(_StringDistances):
    """
    If not, the image's positions and forces are reset to match its centroid.

    Input attributes:
        positions (numpy.ndarray): Atomic positions of the image.
        velocities (numpy.ndarray): Atomic velocities of the image.
        previous_positions (numpy.ndarray): Atomic positions of the image from the previous timestep.
        previous_velocities (numpy.ndarray): Atomic velocities of the image from the previous timestep.
        centroid_positions (numpy.ndarray): The positions of the image's centroid.
        all_centroid_positions (list/numpy.ndarray): A list of positions for all centroids in the string.
        structure (Atoms): The reference structure.
        eps (float): The maximum distance between the closest centroid and the parent centroid to be considered a match
            (i.e. no recentering necessary). (Default is 1e-6.)

    Output attributes:
        positions (numpy.ndarray): Either the original positions passed in, or the previous ones.
        forces (numpy.ndarray): Either the original forces passed in, or the previous ones.
        reflected (bool): Whether or not the image got recentered.
    """
    def __init__(self, name=None):
        super(StringReflect, self).__init__(name=name)

    def command(self, structure, positions, velocities, previous_positions, previous_velocities, centroid_positions,
                all_centroid_positions, eps):
        if self.check_closest_to_parent(structure, positions, centroid_positions, all_centroid_positions, eps):
            return {
                'positions': positions,
                'velocities': velocities,
                'reflected': False
            }
        else:
            return {
                'positions': previous_positions,
                'velocities': -previous_velocities,
                'reflected': True
            }


class PositionsRunningAverage(PrimitiveVertex):
    """
    Calculates the running average of input positions at each call.

    Input attributes:
        positions (list/numpy.ndarray): The instantaneous position, which will be updated to the running average.
        running_average_positions (list/numpy.ndarray): The running average of positions.
        total_steps (int): The total number of times `SphereReflectionPerAtom` is called so far. (Default is 0.)
        thermalization_steps (int): Number of steps the system is thermalized for to reach equilibrium. (Default is
            10 steps.)
        divisor (int): The divisor for the running average positions. Increments by 1, each time the vertex is
            called. (Default is 1.)
        structure (Atoms): The reference structure.

    Output attributes:
        running_average_positions (list/numpy.ndarray): The updated running average list.
        divisor (int): The updated divisor.

    TODO:
        Handle non-static cells, or at least catch them.
        Refactor this so there are list and serial versions equally available
    """

    def __init__(self, name=None):
        super(PositionsRunningAverage, self).__init__(name=name)
        id_ = self.input.default
        id_.total_steps = 0
        id_.thermalization_steps = 10
        id_.divisor = 1

    def command(self, structure, positions, running_average_positions, total_steps, thermalization_steps, divisor):
        total_steps += 1
        if total_steps > thermalization_steps:
            divisor += 1  # On the first step, divide by 2 to average two positions
            weight = 1. / divisor  # How much of the current step to mix into the average
            displacement = structure.find_mic(positions - running_average_positions)
            new_running_average = running_average_positions + (weight * displacement)
            return {
                'running_average_positions': new_running_average,
                'total_steps': total_steps,
                'divisor': divisor,
            }
        else:
            return {
                'running_average_positions': running_average_positions,
                'total_steps': total_steps,
                'divisor': divisor,
            }


class CentroidsRunningAverageMix(PrimitiveVertex):
    """
    Mix in the running average of the positions to the centroid, moving the centroid towards that
        running average by a fraction.

    Input attributes:
        mixing_fraction (float): The fraction of the running average to mix into centroid (Default is 0.1)
        centroids_pos_list (list/numpy.ndarray): List of all the centroids along the string
        running_average_list (list/numpy.ndarray): List of running averages
        structure (Atoms): The reference structure.
        relax_endpoints (bool): Whether or not to relax the endpoints of the string. (Default is False.)

    Output attributes:
        centroids_pos_list (list/numpy.ndarray): List centroids updated towards the running average
    """

    def __init__(self, name=None):
        super(CentroidsRunningAverageMix, self).__init__(name=name)
        self.input.default.mixing_fraction = 0.1
        self.input.default.relax_endpoints = False

    def command(self, structure, mixing_fraction, centroids_pos_list, running_average_positions, relax_endpoints):

        centroids_pos_list = np.array(centroids_pos_list)
        running_average_positions = np.array(running_average_positions)

        updated_centroids = []

        for i, (cent, avg) in enumerate(zip(centroids_pos_list, running_average_positions)):
            if (i == 0 or i == (len(centroids_pos_list) - 1)) and not relax_endpoints:
                updated_centroids.append(cent)
            else:
                displacement = structure.find_mic(avg - cent)
                update = mixing_fraction * displacement
                updated_centroids.append(cent + update)

        return {
            'centroids_pos_list': np.array(updated_centroids)
        }


class CentroidsSmoothing(PrimitiveVertex):
    """
    Global / local smoothing following Vanden-Eijnden and Venturoli (2009). The actual smoothing strength is the
        product of the nominal smoothing strength (`kappa`), the number of images, and the mixing fraction
        (`dtau`).

    Input Attributes:
        kappa (float): Nominal smoothing strength.
        dtau (float): Mixing fraction (from updating the string towards the moving average of the image positions).
        centroids_pos_list (list/numpy.ndarray): List of all the centroid positions along the string.
        structure (Atoms): The reference structure.
        smooth_style (string): Apply 'global' or 'local' smoothing. (Default is 'global'.)

    Output Attributes:
        all_centroid_positions (list/numpy.ndarray): List of smoothed centroid positions.
    """

    def __init__(self, name=None):
        super(CentroidsSmoothing, self).__init__(name=name)
        id_ = self.input.default
        id_.kappa = 0.1
        id_.dtau = 0.1
        id_.smooth_style = 'global'

    def command(self, structure, kappa, dtau, centroids_pos_list, smooth_style):
        n_images = len(centroids_pos_list)
        smoothing_strength = kappa * n_images * dtau
        if smooth_style == 'global':
            smoothing_matrix = self._get_smoothing_matrix(n_images, smoothing_strength)
            smoothed_centroid_positions = np.tensordot(smoothing_matrix, np.array(centroids_pos_list), axes=1)
        elif smooth_style == 'local':
            smoothed_centroid_positions = self._locally_smoothed(smoothing_strength, centroids_pos_list)
        else:
            raise TypeError('Smoothing: choose style = "global" or "local"')
        return {
            'centroids_pos_list': smoothed_centroid_positions
        }

    @staticmethod
    def _get_smoothing_matrix(n_images, smoothing_strength):
        """
        A function that returns the smoothing matrix used in global smoothing.

        Attributes:
            n_images (int): Number of images
            smoothing_strength (float): The smoothing penalty

        Returns:
            smoothing_matrix
        """
        toeplitz_rowcol = np.zeros(n_images)
        toeplitz_rowcol[0] = -2
        toeplitz_rowcol[1] = 1
        second_order_deriv = toeplitz(toeplitz_rowcol, toeplitz_rowcol)
        second_order_deriv[0] = np.zeros(n_images)
        second_order_deriv[-1] = np.zeros(n_images)
        smooth_mat_inv = np.eye(n_images) - smoothing_strength * second_order_deriv

        return np.linalg.inv(smooth_mat_inv)

    @staticmethod
    def _locally_smoothed(structure, smoothing_strength, centroids_pos_list):
        """
        A function that applies local smoothing by taking into account immediate neighbors.

        Attributes:
            structure (Atoms): The reference structure.
            smoothing_strength (float): The smoothing penalty
            centroids_pos_list (list): The list of centroids

        Returns:
            smoothing_matrix
        """
        smoothed_centroid_positions = [centroids_pos_list[0]]
        for i, cent in enumerate(centroids_pos_list[1:-1]):
            left = centroids_pos_list[i]
            right = centroids_pos_list[i+2]
            disp_left = structure.find_mic(cent - left)
            disp_right = structure.find_mic(right - cent)
            switch = (1 + np.cos(np.pi * np.tensordot(disp_left, disp_right) / (
                        np.linalg.norm(disp_left) * (np.linalg.norm(disp_right))))) / 2
            r_star = smoothing_strength * switch * (disp_right - disp_left)
            smoothed_centroid_positions.append(cent + r_star)
        smoothed_centroid_positions.append(centroids_pos_list[-1])

        return smoothed_centroid_positions


class CentroidsReparameterization(PrimitiveVertex):
    """
    Use linear interpolation to equally space the jobs between the first and last job in 3N dimensional space,
        using a piecewise function

    Input attributes:
        centroids_pos_list (list/numpy.ndarray): List of all the centroids along the string
        structure (Atoms): The reference structure.

    Output attributes:
        centroids_pos_list (list/numpy.ndarray): List of equally spaced centroids
    """

    def __init__(self, name=None):
        super(CentroidsReparameterization, self).__init__(name=name)

    def command(self, structure, centroids_pos_list):
        # How long is the piecewise parameterized path to begin with?

        lengths = self._find_lengths(centroids_pos_list, structure)
        length_tot = lengths[-1]
        length_per_frame = length_tot / (len(centroids_pos_list) - 1)

        # Find new positions for the re-parameterized jobs
        new_positions = [centroids_pos_list[0]]
        for n_left, cent in enumerate(centroids_pos_list[1:-1]):
            n = n_left + 1
            length_target = n * length_per_frame

            # Find the last index not in excess of the target length
            try:
                all_not_over = np.argwhere(lengths < length_target)
                highest_not_over = np.amax(all_not_over)
            except ValueError:
                # If all_not_over is empty
                highest_not_over = 0

            # Interpolate from the last position not in excess
            start = centroids_pos_list[highest_not_over]
            end = centroids_pos_list[highest_not_over + 1]
            disp = structure.find_mic(end - start)
            interp_dir = disp / np.linalg.norm(disp)
            interp_mag = length_target - lengths[highest_not_over]

            new_positions.append(start + interp_mag * interp_dir)
        new_positions.append(centroids_pos_list[-1])

        # Apply the new positions all at once
        centroids_pos_list = new_positions

        return {
            'centroids_pos_list': centroids_pos_list
        }

    @staticmethod
    def _find_lengths(a_list, structure):
        """
        Finds the cummulative distance from job to job.

        Attribute:
            a_list (list/numpy.ndarray): List of positions whose lengths are to be calculated
            structure (Atoms): The reference structure.

        Returns:
            lengths (list): Lengths of the positions in the list
        """
        length_cummulative = 0
        lengths = [length_cummulative]
        # First length is zero, all other lengths are wrt the first position in the list
        for n_left, term in enumerate(a_list[1:]):
            disp = structure.find_mic(term - a_list[n_left])
            length_cummulative += np.linalg.norm(disp)
            lengths.append(length_cummulative)
        return lengths



import posixpath

import xml.etree.ElementTree as ET

import numpy as np
from numpy import ndarray
from ase import Atoms as ASEAtoms
from pyiron_base import DataContainer
from pyiron_atomistics import pyiron_to_ase, ase_to_pyiron, Atoms

from pyiron_contrib.atomistics.atomistics.job.structurestorage import StructureStorage
from pyiron_contrib.atomistics.atomicrex.fit_properties import ARFitPropertyList, ARFitProperty
from pyiron_contrib.atomistics.atomicrex.utility_functions import write_pretty_xml
from pyiron_contrib.atomistics.atomicrex.fit_properties import FlattenedARProperty, FlattenedARVectorProperty


class ARStructureContainer:
    def __init__(self):
        self.fit_properties = {}
        # most init can't be done without some information
        # This allows to preallocate arrays and speed everything up massively
        # when writing and reading hdf5 files

    __version__ = "0.2.0"
    __hdf_version__ = "0.2.0"

    def init_structure_container(
        self,
        num_structures,
        num_atoms,
        fit_properties=["atomic-energy", "atomic-forces"],
        structure_file_path=None
        ):
        for p in fit_properties:
            if p == "atomic-forces":
                self.fit_properties[p] = FlattenedARVectorProperty(num_structures=num_structures, num_atoms=num_atoms, prop=p)
            else:
                self.fit_properties[p] = FlattenedARProperty(num_structures=num_structures, prop=p)
        self._init_structure_container(num_structures, num_atoms)
        self.structure_file_path = structure_file_path

    def _init_structure_container(self, num_structures, num_atoms):
        self.flattened_structures = StructureStorage(num_structures=num_structures, num_atoms=num_atoms)
        self.fit = np.empty(num_structures, dtype=bool)
        self.clamp = np.empty(num_structures, dtype=bool)
        self.relative_weight = np.empty(num_structures)

    def add_structure(self, structure, identifier, fit=True, relative_weight=1, clamp=True):
        self.flattened_structures.add_structure(structure, identifier)
        self.fit[self.flattened_structures.prev_structure_index] = fit
        self.relative_weight[self.flattened_structures.prev_structure_index] = relative_weight
        self.clamp[self.flattened_structures.prev_structure_index] = clamp

    def add_scalar_fit_property(
        self,
        prop="atomic-energy",
        target_value=np.nan,
        fit=True,
        relax=False,
        relative_weight=1,
        residual_style=0,
        output=True,
        tolerance=np.nan,
        min_val=np.nan,
        max_val=np.nan,
        ):
        self.fit_properties[prop].target_value[self.flattened_structures.prev_structure_index] = target_value
        self.fit_properties[prop].fit[self.flattened_structures.prev_structure_index] = fit
        self.fit_properties[prop].relax[self.flattened_structures.prev_structure_index] = relax
        self.fit_properties[prop].relative_weight[self.flattened_structures.prev_structure_index] = relative_weight
        self.fit_properties[prop].residual_style[self.flattened_structures.prev_structure_index] = residual_style
        self.fit_properties[prop].output[self.flattened_structures.prev_structure_index] = output
        self.fit_properties[prop].tolerance[self.flattened_structures.prev_structure_index] = tolerance
        self.fit_properties[prop].min_val[self.flattened_structures.prev_structure_index] = min_val
        self.fit_properties[prop].max_val[self.flattened_structures.prev_structure_index] = max_val


    def add_vector_fit_property(
        self,
        prop="atomic-forces",
        target_value=None,
        fit=True,
        relax=False,
        relative_weight=1,
        residual_style=0,
        tolerance=np.nan,
        output=True,
        ):
        if target_value is not None:
            self.fit_properties[prop].target_value[self.flattened_structures.prev_atom_index:self.flattened_structures.current_atom_index] = target_value
        self.fit_properties[prop].fit[self.flattened_structures.prev_structure_index] = fit
        self.fit_properties[prop].relax[self.flattened_structures.prev_structure_index] = relax
        self.fit_properties[prop].relative_weight[self.flattened_structures.prev_structure_index] = relative_weight
        self.fit_properties[prop].residual_style[self.flattened_structures.prev_structure_index] = residual_style
        self.fit_properties[prop].output[self.flattened_structures.prev_structure_index] = output
        self.fit_properties[prop].tolerance[self.flattened_structures.prev_structure_index] = tolerance

    
    def _get_per_structure_index(self, identifier):
        """
        Takes an identifier or an ndarray of indentifiers and returns the corresponding indices in the structure/scalar_fit_properties arrays.
        Args:
            identifiers ([type]): [description]

        Returns:
            [ndarray]: indices corresponding to identifiers in per structure arrays
        """
        if not isinstance(identifier, ndarray):
            identifier = np.array(identifier)
        indices = np.flatnonzero(
            np.isin(
                self.flattened_structures._per_structure_arrays["identifier"], identifier, assume_unique=True
                )
            )
        # This is some sorting magic that could lead to strange errors
        # Look here if something goes wrong with the order of
        ids_stored = np.array(self.flattened_structures._per_structure_arrays["identifier"][indices])
        sorter = ids_stored.argsort()[identifier.argsort()]
        return indices[sorter]

    def get_scalar_property(self, prop, identifier, final=True):
        """
        Returns final or target value of a scalar property based on the identifier used when adding a structure
        If identifier is an array of identifiers the return value is an array of values.

        Args:
            prop (str): property string as used when adding the property
            identifier (str or array(str)): identifier as used when adding the structure
            final (bool, optional): Whether to return the final or target value. Defaults to True.

        Returns:
            [type]: [description]
        """
        index = self._get_per_structure_index(identifier)
        if final:
            return self.fit_properties[prop].final_value[index]
        else:
            return self.fit_properties[prop].target_value[index]

    def get_vector_property(self, prop, identifier, final=True):
        """
        Returns final or target value of a vector property based on the identifier used when adding a structure.
        Currently only allows to return a vector property for a single structure, not for multiple ones like the
        get_scalar_property function.

        Args:
            prop (str): property string as used when adding the property
            identifier (str): identifier as used when adding the structure
            final (bool, optional): Whether to return the final or target value. Defaults to True.

        Returns:
            [type]: [description]
        """
        if not isinstance(identifier, str):
            raise NotImplementedError("Can only look up properties for single identifiers currently")
        index = self._get_per_structure_index(identifier)[0]
        slc  = self.flattened_structures._get_per_atom_slice(index)
        if final:
            return self.fit_properties[prop].final_value[slc]
        else:
            return self.fit_properties[prop].target_value[slc]


    def _type_to_hdf(self, hdf):
        """
        Internal helper function to save type and version in hdf root

        Args:
            hdf (ProjectHDFio): HDF5 group object
        """
        hdf["NAME"] = self.__class__.__name__
        hdf["TYPE"] = str(type(self))
        hdf["VERSION"] = self.__version__
        hdf["HDF_VERSION"] = self.__hdf_version__

    def to_hdf(self, hdf, group_name="structures"):
        with hdf.open(group_name) as h:
            self._type_to_hdf(h)
            self.flattened_structures.to_hdf(hdf=h)

            h_fit = h.create_group("fit_properties")
            for k, v in self.fit_properties.items():
                v.to_hdf(h_fit, group_name=k)
            
            h["fit"] = self.fit
            h["clamp"] = self.clamp
            h["relative_weight"] = self.relative_weight
            h["structure_file_path"] = self.structure_file_path

    def from_hdf(self, hdf, group_name="structures"):
        with hdf.open(group_name) as h:
            version = h.get("HDF_VERSION", "0.1.0")
            # Compatibility Old and new StructureStorage
            if version == "0.1.0":
                num_structures = h["flattened_structures/num_structures"]
                num_atoms = h["flattened_structures/num_atoms"]
                group_name_2 = "flattened_structures"
            else:
                num_structures = h["structures/num_structures"]
                num_atoms = h["structures/num_atoms"]
                group_name_2 = "structures"

            self._init_structure_container(num_structures, num_atoms)
            self.flattened_structures.from_hdf(hdf=h, group_name=group_name_2)

            h_fit = h["fit_properties"]
            for group in h_fit.list_groups():
                if group == "atomic-forces":
                    self.fit_properties[group] = FlattenedARVectorProperty(num_structures, num_atoms, group)
                    self.fit_properties[group].from_hdf(h_fit, group_name=group)
                else:
                    self.fit_properties[group] = FlattenedARProperty(num_structures, group)
                    self.fit_properties[group].from_hdf(h_fit, group_name=group)

            self.clamp = h["clamp"]
            self.fit = h["fit"]
            self.relative_weight = h["relative_weight"]
            self.structure_file_path = h["structure_file_path"]
    

    def write_xml_file(self, directory, name="structures.xml"):
        """
        Internal helper function that writes an atomicrex style
        xml file containg all structures.

        Args:
            directory (string): Working directory.
            name (str, optional): . Defaults to "structures.xml".
        """        
        root = ET.Element("group")
        if self.structure_file_path is None:

            # write POSCAR and xml
            for i in range(self.flattened_structures.num_structures):
                vec_start = self.flattened_structures.start_index[i]
                vec_end = self.flattened_structures.start_index[i]+self.flattened_structures.length[i]
                write_modified_poscar(
                    identifier=self.flattened_structures.identifier[i],
                    forces=self.fit_properties["atomic-forces"].target_value[vec_start:vec_end],
                    positions=self.flattened_structures.positions[vec_start:vec_end],
                    symbols=self.flattened_structures.symbols[vec_start:vec_end],
                    cell=self.flattened_structures.cell[i],
                    directory=directory
                )
                
                fit_properties_xml = ET.Element("properties")
                for flat_prop in self.fit_properties.values():
                    fit_properties_xml.append(flat_prop.to_xml_element(i))

                struct_xml = structure_meta_xml(
                    identifier=self.flattened_structures.identifier[i],
                    relative_weight=self.relative_weight[i],
                    clamp=self.clamp[i],
                    fit_properties=fit_properties_xml,
                    struct_file_path=self.structure_file_path,
                    fit = self.fit[i],
                )
                root.append(struct_xml)
        else:
            # write only xml and use POSCARs written already to some path
            for i in range(self.flattened_structures.num_structures):
                fit_properties_xml = ET.Element("properties")
                for flat_prop in self.fit_properties.values():
                    fit_properties_xml.append(flat_prop.to_xml_element(i))

                struct_xml = structure_meta_xml(
                    identifier=self.flattened_structures.identifier[i],
                    relative_weight=self.relative_weight[i],
                    clamp=self.clamp[i],
                    fit_properties=fit_properties_xml,
                    struct_file_path=self.structure_file_path,
                    fit = self.fit[i],
                )
                root.append(struct_xml)
        filename = posixpath.join(directory, name)
        write_pretty_xml(root, filename)


    def _parse_final_properties(self, struct_lines):
        """
        Internal function that parses the values of fitted properties
        calculated with the final iteration of the fitted potential.

        Args:
            struct_lines (list[str]): lines from atomicrex output that contain structure information.
        """ 
        force_vec_triggered = False
        for l in struct_lines:
            l = l.strip()
            
            if force_vec_triggered:
                if l.startswith("atomic-forces:"):
                    l = l.split()
                    index = int(l[1])
                    final_forces[index, 0] = float(l[2].lstrip("(").rstrip(","))
                    final_forces[index, 1] = float(l[3].rstrip(","))
                    final_forces[index, 2] = float(l[4].rstrip(")"))
                else:
                    force_vec_triggered = False
                    start_index = self.flattened_structures.start_index[s_index]
                    self.fit_properties["atomic-forces"].final_value[start_index:start_index+len_struct] = final_forces

            # This has to be if and not else because it has to run in the same iteration. Empty lines get skipped.
            if not force_vec_triggered and l:
                if l.startswith("Structure"):
                    s_id = l.split("'")[1]
                    s_index = np.nonzero(self.flattened_structures.identifier==s_id)[0][0]

                else:
                    if not l.startswith("atomic-forces avg/max:"):
                        prop, f_val = ARFitProperty._parse_final_value(line=l)
                        if prop in self.fit_properties.keys():
                            self.fit_properties[prop].final_value[s_index] = f_val
                    else:
                        force_vec_triggered = True
                        len_struct = self.flattened_structures.length[s_index]
                        final_forces = np.empty((len_struct, 3))
    

### This is probably useless like this in most cases because forces can't be passed.
def structure_to_xml_element(structure):
    """
    Converts an ase/pyiron atoms object to an atomicrex xml element
    Right now forces can't be passed in the xml file, so this is not really helpful.
    Args:
        structure (Atoms): ase or pyiron Atoms

    Returns:
        (ET.Element, ET.Element): atomicrex structure xml
    """    
    pbc = ET.Element("pbc")
    pbc.set("x", f"{structure.pbc[0]}".lower())
    pbc.set("y", f"{structure.pbc[1]}".lower())
    pbc.set("z", f"{structure.pbc[2]}".lower())

    c = structure.cell
    cell = ET.Element("cell")
    for i in range(3):
        a = ET.SubElement(cell, f"a{i+1}")
        a.set("x", f"{c[i][0]}")
        a.set("y", f"{c[i][1]}")
        a.set("z", f"{c[i][2]}")

    atoms = ET.SubElement(cell, "atoms")
    for at in structure:
        a = ET.SubElement(atoms, "atom")
        a.set("type", at.symbol)
        a.set("x", f"{at.a}")
        a.set("y", f"{at.b}")
        a.set("z", f"{at.c}")
        a.set("reduced", "false")

    return pbc, cell


class ARStructure(object):
    """
    Class that contains a pyiron structure, an ARFitPropertyList
    and some attributes that define the fitting procedure.
    Provides internal helper methods.
    """

    def __init__(self, structure=None, fit_properties=None, identifier=None, relative_weight=1, clamp=True, fit=True):
        self._structure = None
        self._fit_properties = None
        if structure is not None:
            self.structure = structure
        if fit_properties is not None:
            self.fit_properties = fit_properties
        self.identifier = identifier
        self.relative_weight = relative_weight
        self.clamp = clamp
        self.fit = fit

    @property
    def structure(self):
        return self._structure

    @structure.setter
    def structure(self, structure):
        if not isinstance(structure, Atoms):
            raise ValueError("structure has to be a pyiron atoms instance")
        self._structure = structure

    @property
    def fit_properties(self):
        return self._fit_properties

    @fit_properties.setter
    def fit_properties(self, fit_properties=ARFitPropertyList()):
        if not isinstance(fit_properties, ARFitPropertyList):
            raise ValueError("fit_properties have to be an ARFitPropertyList")
        self._fit_properties = fit_properties

    def _write_poscar_return_xml(self, directory, struct_file_path):
        """
        Internal function that writes the structure in an extended POSCAR
        format that contains forces. Returns an atomicrex xml
        element that references the POSCAR file and contains additional
        information like the fit properties.

        Args:
            directory (string): Working directory
            struct_file_path (string): path to a directory containing structure files. If None structure files are written in working directory
        Returns:
            [ElementTree xml element]: xml element referencing POSCAR, contains additional information
        """        
        if struct_file_path is None:
            forces = self.fit_properties["atomic-forces"].target_value
            write_modified_poscar(
                identifier = self.identifier,
                structure = self.structure,
                forces=forces,
                directory=directory
                )

        return structure_meta_xml(
            identifier = self.identifier,
            relative_weight = self.relative_weight,
            clamp = self.clamp,
            fit_properties = self.fit_properties,
            fit = self.fit,
            struct_file_path = struct_file_path,
        )

    def to_hdf(self, hdf=None, group_name="arstructure", full_hdf=True):
        """
        Internal function
        """        
        with hdf.open(group_name) as hdf_s_lst:
            if full_hdf:
                self.structure.to_hdf(hdf=hdf_s_lst, group_name="structure")
            self.fit_properties.to_hdf(hdf=hdf_s_lst, group_name="fit_properties")
            hdf_s_lst["relative_weight"] = self.relative_weight
            hdf_s_lst["identifier"] = self.identifier
            hdf_s_lst["clamp"] = self.clamp

    def from_hdf(self, hdf=None, group_name="arstructure", full_hdf=True):
        """
        Internal function 
        """    
        with hdf.open(group_name) as hdf_s_lst:
            if full_hdf:
                structure = Atoms()
                structure.from_hdf(hdf_s_lst, group_name="structure")
                self.structure = structure
            self.fit_properties = ARFitPropertyList()
            self.fit_properties.from_hdf(hdf=hdf_s_lst, group_name="fit_properties")
            self.relative_weight = hdf_s_lst["relative_weight"]
            self.identifier = hdf_s_lst["identifier"]
            self.clamp = hdf_s_lst["clamp"]


class ARStructureList(object):
    """
    Container class for AR structures. structures attribute
    of the atomicrex job class.
    Provides functions for internal use and a convenient way
    to add additional structures to the atomicrex job.

    Will be replaced with FlattenedStructureContainer or used only for representation of the data
    """    
    def __init__(self):
        self._structure_dict = {}

        # This allows to give a path where structures are stored in atomicrex POSCAR format
        # It modifies the write input function to prevent a lot of file io and slow down performance
        # If it is not given a file is written for every structure of the job.
        self.struct_file_path = None
        self.full_structure_to_hdf = True
        self.num_structures = 0
        self.num_atoms = 0

    def add_structure(self, structure, identifier, fit_properties=None, relative_weight=1, clamp=True, fit=True):
        """
        Provides a convenient way to add additional
        structures to the job.
        
        Args:
            structure (Pyiron Atoms): structure that should be added.
            identifier (string): ID string. Must be unique or overwrites the old structure.
            fit_properties ([ARFitPropertyList], optional): Fit Properties can be conveniently added after appending the structure. Defaults to None.
            relative_weight (int, optional): Assigns a weight for the objective function. Defaults to 1.
            clamp (bool, optional): Clamp the structure (Do not relax it). Defaults to True.

        Raises:
            ValueError: Raises if structure is not a pyiron atoms instance.

        Returns:
            [ARStructure]: Acces to the atomicrex structure in the structure list.
        """        
        if isinstance(structure, Atoms):
            if fit_properties is None:
                fit_properties = ARFitPropertyList()
            ar_struct = ARStructure(structure, fit_properties, identifier, relative_weight=relative_weight, clamp=clamp)
            self.num_structures += 1
            self.num_atoms += len(structure)
            return self._add_ARstructure(ar_struct)
        else:
            raise ValueError("Structure has to be a Pyiron Atoms instance")


    def _add_ARstructure(self, structure):
        """Internal helper function to be able to implement some more checks later on
        """        
        identifier = f"{structure.identifier}"
        self._structure_dict[identifier] = structure
        return self._structure_dict[identifier]

    def write_xml_file(self, directory, name="structures.xml"):
        """
        Internal helper function that write an atomicrex style
        xml file containg all structures.

        Args:
            directory (string): Working directory.
            name (str, optional): . Defaults to "structures.xml".
        """        
        root = ET.Element("group")
        for s in self._structure_dict.values():
            root.append(s._write_poscar_return_xml(directory, self.struct_file_path))
        filename = posixpath.join(directory, name)
        write_pretty_xml(root, filename)


    #def to_hdf(self, hdf=None, group_name="arstructurelist"):

    #def from_hdf(self, hdf=None, group_name="arstructurelist"):
    
    def to_hdf(self, hdf=None, group_name="arstructurelist"):
        """
        Internal function 
        """    
        with hdf.open(group_name) as hdf_s_lst:
            hdf_s_lst["struct_file_path"] = self.struct_file_path
            hdf_s_lst["full_structure_to_hdf"] = self.full_structure_to_hdf
            for k, v in self._structure_dict.items():
                v.to_hdf(hdf=hdf_s_lst, group_name=k, full_hdf=self.full_structure_to_hdf)

    def from_hdf(self, hdf=None, group_name="arstructurelist"):
        """
        Internal function 
        """        
        with hdf.open(group_name) as hdf_s_lst:
            # compatibility with my old jobs. find a way to write this to hdf for them and delete try except block
            try:
                self.struct_file_path = hdf_s_lst["struct_file_path"]
                self.full_structure_to_hdf = hdf_s_lst["full_structure_to_hdf"]
            except ValueError:
                pass

            for g in sorted(hdf_s_lst.list_groups()):
                s = ARStructure()
                s.from_hdf(hdf=hdf_s_lst, group_name=g, full_hdf=self.full_structure_to_hdf)
                self._structure_dict[g] = s


    def _parse_final_properties(self, struct_lines):
        """
        Internal function that parses the values of fitted properties
        calculated with the final iteration of the fitted potential.

        Args:
            struct_lines (list[str]): lines from atomicrex output that contain structure information.
        """ 
        force_vec_triggered = False   
        for l in struct_lines:
            l = l.strip()
            
            if force_vec_triggered:
                if l.startswith("atomic-forces:"):
                    l = l.split()
                    index = int(l[1])
                    final_forces[index, 0] = float(l[2].lstrip("(").rstrip(","))
                    final_forces[index, 1] = float(l[3].rstrip(","))
                    final_forces[index, 2] = float(l[4].rstrip(")"))
                else:
                    force_vec_triggered = False
                    s.fit_properties["atomic-forces"].final_value = final_forces
            
            # This has to be if and not else because it has to run in the same iteration. Empty lines get skipped.
            if not force_vec_triggered and l:
                if l.startswith("Structure"):
                    s_id = l.split("'")[1]
                    s = self._structure_dict[s_id]
            
                else:
                    if not l.startswith("atomic-forces avg/max:"):
                        prop, f_val = ARFitProperty._parse_final_value(line=l)
                        if prop in s.fit_properties:
                            s.fit_properties[prop].final_value = f_val
                    else:
                        force_vec_triggered = True
                        final_forces = np.empty((len(s.structure), 3))



def write_modified_poscar(identifier, forces, directory, structure=None, positions=None, cell=None, symbols=None):
    """
    Internal function. Writes a pyiron structure
    and corresponding forces in a modified POSCAR file.
    Either provide a structure instance or positions and symbols.
    Args:
        identifier (str): Unique identifier used for the filename.
        structure (Atoms): ase or pyiron atoms object. 
        positions (np.array): atomic positions
        symbols (np.array): chemical symbols
        forces (array or list[list]): atomic forces. Must be in same order as positions.
        directory (str): Working directory.
    """    
    filename = posixpath.join(directory, f"POSCAR_{identifier}")
    with open(filename, 'w') as f:
        # Elements as system name
        # Also check if the symbols are disordered
        # to prevent errors resulting from poscar format
        # not having indices or element for each atom.
        elements = []
        counter = 0
        last_elem = "XX"
        if symbols is None:
            symbols = np.array(structure.symbols)
        for elem in symbols:
            if not elem in elements:
                elements.append(elem)
            if not last_elem == elem:
                counter += 1
                last_elem = elem
        if counter != len(elements):
            raise ValueError(
                "Structures with disordered elements are not supported right now\n"
                "They can be sorted together with target forces using numpy.argsort"
            )
            ## maybe this can be fixed returning an argsort array to sort the forces.
        f.write(f"{elements}\n")

        # Cell metric
        f.write('1.0\n')
        if cell is None:
            cell = structure.cell
        for vec in cell:
            for v in vec:
                f.write(f'{v} ')
            f.write('\n')

        # Element names
        for elem in elements:
            f.write(f'{elem} ')
        f.write('\n')

        # Number of elements per type
        for elem in elements:
            n = len(symbols[symbols==elem])
            f.write(f'{n}    ')
        f.write('\n')

        # Scaled coordinates
        if positions is None:
            positions = structure.positions
        f.write('Cartesian\n')

        # Coordinates and forces
        for pos, force in zip(positions, forces):
            for r in pos: f.write((f'{r} '))
            f.write('   ')
            for r in force: f.write((f'{r} '))
            f.write('\n')


def structure_meta_xml(
        identifier,
        relative_weight,
        clamp, ## Not sure if and how this combines with relax in the ARFitParameter sets
        fit_properties,
        struct_file_path,
        fit,
        mod_poscar = True,
        
):
    """
    Internal function. Creates xml element with
    scalar properties, weight and reference to POSCAR
    containg structure and forces.

    Args:
        identifier (str): Unique identifier.
        relative_weight (float): weight in objective function
        clamp (bool): clamp the structure (no relaxation) 
        mod_poscar (bool, optional): Combine with poscar file. Defaults to True.
        struct_file_path (string): path to POSCAR file
    Raises:
        NotImplementedError: mod_poscar has to be True, because forces can't
        be provided in xml format in atomicrex yet.

    Returns:
        [ElementTree xml element]: atomicrex structure xml element.
    """

    struct_xml = ET.Element("user-structure")
    struct_xml.set("id", f"{identifier}")
    struct_xml.set("relative-weight", f"{relative_weight}")
    if fit: 
        struct_xml.set("fit", "true")
    else:
        struct_xml.set("fit", "false")

    if mod_poscar:
        poscar_file = ET.SubElement(struct_xml, "poscar-file")
        if struct_file_path is None:
            poscar_file.text = f"POSCAR_{identifier}"
        else:
            poscar_file.text = f"{struct_file_path}POSCAR_{identifier}"
    else:
        #    struct_xml.extend(structure_to_xml_element(structure))
        raise NotImplementedError(
            "Only writing structure meta information, \n"
            "not structure data for now because this is not implemented in atomicrex"
        )

    if not clamp:
        relax_dof = ET.SubElement(struct_xml, "relax-dof")
        atom_coordinates = ET.SubElement(relax_dof, "atom-coordinates")
    
    if isinstance(fit_properties, ARFitPropertyList):
        properties = ET.SubElement(struct_xml, "properties")
        for prop in fit_properties.values():
            properties.append(prop.to_xml_element())
    else:
        struct_xml.append(fit_properties)

    return struct_xml
